@@ -33,6 +33,7 @@ def ignore_filter_from_tmpl(template, ucr_key, default=''):
     '(|(cn=one)(cn=two)(cn=three))'
     """
     from univention.connector.ad import format_escaped
+
     variables = [v for v in configRegistry.get(ucr_key, default).split(',') if v]
     filter_parts = [format_escaped(template, v) for v in variables]
     if filter_parts:
@@ -176,6 +177,7 @@ def create_mapping(configbasename='connector'):
             post_con_modify_functions=list(filter(None, [
                 univention.connector.ad.set_userPrincipalName_from_ucr,
                 univention.connector.ad.password.password_sync_ucs if configRegistry.is_false(connector('%s/ad/mapping/user/password/disabled'), True) else None,
+                univention.connector.ad.password.lockout_sync_from_ucs,
                 univention.connector.ad.primary_group_sync_from_ucs,
                 univention.connector.ad.object_memberships_sync_from_ucs,
                 univention.connector.ad.disable_user_from_ucs,
@@ -183,6 +185,7 @@ def create_mapping(configbasename='connector'):
             post_ucs_modify_functions=list(filter(None, [
                 univention.connector.ad.password.password_sync_kinit if configRegistry.is_false(connector('%s/ad/mapping/user/password/disabled'), True) and configRegistry.is_true(connector('%s/ad/mapping/user/password/kinit'), False) else None,
                 univention.connector.ad.password.password_sync if configRegistry.is_false(connector('%s/ad/mapping/user/password/disabled'), True) and not configRegistry.is_true(connector('%s/ad/mapping/user/password/kinit'), False) else None,
+                univention.connector.ad.password.lockout_sync_to_ucs,
                 univention.connector.ad.set_univentionObjectFlag_to_synced,
                 univention.connector.ad.primary_group_sync_to_ucs,
                 univention.connector.ad.object_memberships_sync_to_ucs,
@@ -493,7 +496,7 @@ def create_mapping(configbasename='connector'):
                 ),
             },
         ),
-    }
+    }  # fmt: skip
 
     # allow filter
     for obj_type in ad_mapping.keys():  # noqa: PLC0206
