@@ -3,35 +3,8 @@
 # Univention AD Connector
 #  Resync object from AD to OpenLDAP
 #
-# Like what you see? Join us!
-# https://www.univention.com/about-us/careers/vacancies/
-#
-# Copyright 2018-2024 Univention GmbH
-#
-# https://www.univention.de/
-#
-# All rights reserved.
-#
-# The source code of this program is made available
-# under the terms of the GNU Affero General Public License version 3
-# (GNU AGPL V3) as published by the Free Software Foundation.
-#
-# Binary versions of this program provided by Univention to you as
-# well as other copyrighted, protected or trademarked materials like
-# Logos, graphics, fonts, specific documentations and configurations,
-# cryptographic keys etc. are subject to a license agreement between
-# you and Univention and not subject to the GNU AGPL V3.
-#
-# In the case you use this program under the terms of the GNU AGPL V3,
-# the program is provided in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-# GNU Affero General Public License for more details.
-#
-# You should have received a copy of the GNU Affero General Public
-# License with the Debian GNU/Linux or Univention distribution in file
-# /usr/share/common-licenses/AGPL-3; if not, see
-# <https://www.gnu.org/licenses/>.
+# SPDX-FileCopyrightText: 2018-2025 Univention GmbH
+# SPDX-License-Identifier: AGPL-3.0-only
 
 
 import os
@@ -46,6 +19,7 @@ from samba.ndr import ndr_unpack
 
 import univention.connector.ad
 from univention.config_registry import ConfigRegistry
+from univention.dn import DN
 
 
 class GUIDNotFound(BaseException):
@@ -59,7 +33,7 @@ class DNNotFound(BaseException):
 class ad(univention.connector.ad.ad):
 
     def _remove_cache_entries(self, guid):
-        cache_filename = '/etc/univention/%s/adcache.sqlite' % CONFIGBASENAME
+        cache_filename = f'/etc/univention/{CONFIGBASENAME}/adcache.sqlite'
         if not os.path.exists(cache_filename):
             return
         cache_db = sqlite3.connect(cache_filename)
@@ -75,7 +49,7 @@ class ad(univention.connector.ad.ad):
         os.chmod(cache_filename, 640)
 
     def _add_object_to_rejected(self, ad_dn, usn):
-        state_filename = '/etc/univention/%s/internal.sqlite' % CONFIGBASENAME
+        state_filename = f'/etc/univention/{CONFIGBASENAME}/internal.sqlite'
         db = sqlite3.connect(state_filename)
         c = db.cursor()
         c.execute("INSERT OR REPLACE INTO 'AD rejected' (key, value) VALUES (?, ?);", (usn, ad_dn))
@@ -84,8 +58,13 @@ class ad(univention.connector.ad.ad):
         os.chmod(state_filename, 640)
 
     def resync(self, ad_dns=None, ldapfilter=None, ldapbase=None):
+        result = self.search_ad(ad_dns, ldapfilter, ldapbase)
+
+        # If a DN to resync happens to be a subtree DN, we might also want to resync the ancestors
+        self.prepend_ancestors_of_allowed_subtrees(result, ldapfilter, ldapbase)
+
         treated_dns = []
-        for ad_dn, guid, usn in self.search_ad(ad_dns, ldapfilter, ldapbase):
+        for ad_dn, guid, usn in result:
             self._remove_cache_entries(guid)
             self._add_object_to_rejected(ad_dn, usn)
             treated_dns.append(ad_dn)
@@ -127,7 +106,7 @@ class ad(univention.connector.ad.ad):
                 ldapfilter = '(objectClass=*)'
 
             if not ldapbase:
-                ldapbase = self.configRegistry['%s/ad/ldap/base' % CONFIGBASENAME]
+                ldapbase = self.configRegistry[f'{CONFIGBASENAME}/ad/ldap/base']
 
             guid = None
             try:
@@ -148,6 +127,45 @@ class ad(univention.connector.ad.ad):
 
         return search_result
 
+    def _get_allowed_subtrees(self) -> list[DN]:
+        allowed_subtrees: list[DN] = []
+
+        for key in self.configRegistry:
+            if key.startswith(f'{CONFIGBASENAME}/ad/mapping/allowsubtree') and key.endswith('/ad'):
+                allowed_subtrees.append(DN(self.configRegistry[key]))
+
+        return allowed_subtrees
+
+    def prepend_ancestors_of_allowed_subtrees(self, ad_search_result: list[tuple] | None, ldapfilter, ldapbase):
+        if self.configRegistry.is_false(f"{CONFIGBASENAME}/ad/mapping/allow-subtree-ancestors", False) or ad_search_result is None:
+            return
+
+        ad_ldap_base = DN(self.configRegistry.get("connector/ad/ldap/base"))
+
+        allowed_subtrees = self._get_allowed_subtrees()
+
+        ancestor_list = []
+        for object_dn, _, _ in ad_search_result:
+            object_dn = DN(object_dn)
+            if object_dn not in allowed_subtrees:
+                continue
+
+            subtree_dn = object_dn
+            parent_dn = object_dn.parent
+            while parent_dn and parent_dn != ad_ldap_base:
+                parent = str(parent_dn)
+                if parent not in ancestor_list:
+                    print(f"{subtree_dn} is an allowed subtree. Adding ancestor DN {parent} to resync list.")
+                    ancestor_list.insert(0, parent)
+                parent_dn = parent_dn.parent
+
+        if not ancestor_list:
+            return
+
+        ancestor_ad_search_result = self.search_ad(ancestor_list, ldapfilter, ldapbase)
+        ancestor_ad_search_result.sort(key=lambda x: len(x[0]))
+        ad_search_result[:0] = ancestor_ad_search_result
+
 
 if __name__ == '__main__':
     parser = ArgumentParser(description="Resync object from AD to UCS")
@@ -158,9 +176,9 @@ if __name__ == '__main__':
     options = parser.parse_args()
 
     CONFIGBASENAME = options.configbasename
-    state_directory = '/etc/univention/%s' % CONFIGBASENAME
+    state_directory = f'/etc/univention/{CONFIGBASENAME}'
     if not os.path.exists(state_directory):
-        parser.error("Invalid configbasename, directory %s does not exist" % state_directory)
+        parser.error(f"Invalid configbasename, directory {state_directory} does not exist")
 
     if not options.dn and not options.ldapfilter:
         parser.print_help()
@@ -169,7 +187,7 @@ if __name__ == '__main__':
     configRegistry = ConfigRegistry()
     configRegistry.load()
 
-    poll_sleep = int(configRegistry['%s/ad/poll/sleep' % CONFIGBASENAME])
+    poll_sleep = int(configRegistry[f'{CONFIGBASENAME}/ad/poll/sleep'])
     ad_init = None
 
     ad_dns = list(filter(None, [options.dn]))
@@ -185,27 +203,27 @@ if __name__ == '__main__':
         sys.stdout.flush()
         time.sleep(poll_sleep)
     except DNNotFound as ex:
-        print('ERROR: The AD object was not found: %s' % (ex.args[1],))
+        print(f'ERROR: The AD object was not found: {ex.args[1]}')
         if len(ex.args) == 3:
             treated_dns = ex.args[2]
         sys.exit(1)
     except GUIDNotFound as ex:
-        print('ERROR: The AD search for objectGUID failed: %s' % (ex.args[1],))
+        print(f'ERROR: The AD search for objectGUID failed: {ex.args[1]}')
         if len(ex.args) == 3:
             treated_dns = ex.args[2]
         sys.exit(1)
     finally:
         for dn in treated_dns:
-            print('resync triggered for %s' % dn)
+            print(f'resync triggered for {dn}')
 
     if treated_dns:
         estimated_delay = 60
         try:
-            estimated_delay = int(resync.configRegistry.get('%s/ad/retryrejected' % CONFIGBASENAME, 10)) * int(resync.configRegistry.get('%s/ad/poll/sleep' % CONFIGBASENAME, 5))
+            estimated_delay = int(resync.configRegistry.get(f'{CONFIGBASENAME}/ad/retryrejected', 10)) * int(resync.configRegistry.get(f'{CONFIGBASENAME}/ad/poll/sleep', 5))
         except ValueError:
             pass
 
-        print('Estimated sync in %s seconds.' % (estimated_delay,))
+        print(f'Estimated sync in {estimated_delay} seconds.')
     else:
         print('No matching objects.')
 
